# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
import subprocess

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, MissingFileError

class RTTGetter_WorkerThread(BaseConfigEnricherThread):

    DESCR = "RTTGetter"

    # The following regex pattern is reported by docs/RTT_GETTER.rst
    RETURN_VALUE_RE_PATTERN = re.compile("^\d+[.]?\d*$")

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.rtt_getter_path = None

    @staticmethod
    def _parse_result(raw):
        if not raw:
            raise ValueError("no value returned")

        res = raw.strip()
        if not res:
            raise ValueError("empty value returned")

        res = res.split("\n")[0]
        if res.lower() == "none":
            return None

        if not RTTGetter_WorkerThread.RETURN_VALUE_RE_PATTERN.match(res):
            raise ValueError("invalid value: {}".format(res))

        return float(res)

    def do_task(self, task):
        client = task

        cmd = [self.rtt_getter_path]
        cmd += [client["ip"]]
        cmd += [str(client["asn"])]
        cmd += [str(client["id"])]

        try:
            out = subprocess.check_output(cmd)
        except Exception as e:
            err = "Error while executing RTT getter command '{}': {}".format(
                " ".join(cmd), str(e)
            )
            logging.error(err)
            raise BuilderError()

        try:
            return self._parse_result(out.decode("utf-8"))
        except ValueError as e:
            err = ("Error while parsing result from "
                   "RTT getter command '{}': {}".format(" ".join(cmd), str(e)))
            logging.error(err)
            raise BuilderError()

    def save_data(self, task, data):
        client = task
        rtt = data
        client["rtt"] = rtt

class RTTGetterConfigEnricher(BaseConfigEnricher):

    WORKER_THREAD_CLASS = RTTGetter_WorkerThread

    def prepare(self):
        path = self.builder.rtt_getter_path
        if path:
            if not os.path.exists(path):
                raise MissingFileError(path)
            if not (os.path.isfile(path) and \
                    os.access(path, os.X_OK)):
                raise BuilderError(
                    "The file {} used for rtt_getter_path is not "
                    "executable.".format(path)
                )
        else:
            raise BuilderError("Path of the RTT getter program is missing.")

    def _config_thread(self, thread):
        thread.rtt_getter_path = self.builder.rtt_getter_path

    def add_tasks(self):
        # Enqueuing tasks.
        for client in self.builder.cfg_clients.cfg["clients"]:
            self.tasks_q.put(client)



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .irr_db_dump import GenericIRRWhoisDBDumpEnricher
from ..arin_db_dump import ARINWhoisDBDump

class ARINWhoisDBDumpEnricher(GenericIRRWhoisDBDumpEnricher):

    DESCR = "ARIN"
    DIR_NAME = "arin_db"
    CONFIG_SECTION_NAME = "use_arin_bulk_whois_data"
    PARSER_CLASS = ARINWhoisDBDump
    BUILDER_TARGET_DICT_NAME = "arin_whois_records"



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import yaml

from pierky.arouteserver.errors import ConfigError, ARouteServerError
from pierky.arouteserver.tests.base import ARouteServerTestCase


class TestConfigParserBase(ARouteServerTestCase):

    NEED_TO_CAPTURE_LOG = True

    FILE_PATH = None
    CONFIG_PARSER_CLASS = None

    def _setUp(self):
        self.load_config()

    def load_config(self, file_name=None, yaml=None):
        self.cfg = None
        self.cfg = self.CONFIG_PARSER_CLASS()
        if yaml:
            self.cfg._load_from_yaml(yaml)
        elif file_name:
            self.cfg.load(file_name)
        else:
            self.cfg.load(self.FILE_PATH)

    def _contains_err(self, err=None):
        logging.info("-----------------------------------------------")
        self.clear_log()

        exception_raised = False
        try:
            self.cfg.parse()
        except ARouteServerError as e:
            exception_raised = True
            if err is None:
                self.fail(
                    "An unexpected exception occurred: {}".format(
                        str(e)
                    )
                )
        if err is not None and not exception_raised:
            self.fail(
                "An error was expected but no exceptions have been raised"
            )

        res = self.logger_handler.msgs

        found = False

        if err:
            for s in res:
                if err in s:
                    found = True

            if not found:
                self.fail("Error message expected but not found:\n"
                          " - expected:\n"
                          "\t{}\n"
                          " - received: \n"
                          "\t{}".format(
                              err, "\n\t".join(res)
                              ))
        elif res:
            raise self.fail(
                "Unexpected error message(s) found:\n"
                "\t{}".format("\n\t".join(res))
            )

    def _test_mandatory(self, dic, key, has_default=False):
        del dic[key]
        if has_default:
            self._contains_err()
        else:
            self._contains_err("Can't be empty.")

        for v in (None, "", "     "):
            dic[key] = v
            if has_default:
                self._contains_err()
            else:
                self._contains_err("Can't be empty.")

    def _test_optional(self, dic, key):
        del dic[key]
        self._contains_err()

        for v in (None, "", "     "):
            dic[key] = v
            self._contains_err()

    def _test_bool_val(self, dic, key):
        for v in (True, "true", "True", "tRuE", "yes", "YES", "t", "T", "1", 1):
            dic[key] = v
            self._contains_err()
        for v in (False, "false", "False", "fAlSe", "no", "NO", "f", "F", "0", 0):
            dic[key] = v
            self._contains_err()
        for v in ("sure"):
            dic[key] = v
            self._contains_err("Invalid boolean value: {}.".format(v))

    def _test_option(self, dic, key, valid_options):
        for v in valid_options:
            dic[key] = v
            self._contains_err()
        for v in ("xxx", "yyy"):
            dic[key] = v
            self._contains_err("Invalid option for '{}': '{}';".format(key, v))

    def _test_ip_min_max_len(self, dic, key, min_max, ip_ver, valid, invalid):
        for v in valid:
            dic[key][min_max] = v
            self._contains_err()
        for v in invalid:
            dic[key][min_max] = v
            self._contains_err("'{}' in the IPv{} min/max length".format(min_max, ip_ver))



import os
from os.path import abspath, dirname, join
from setuptools import setup, find_packages

"""
New release procedure

- ./utils/update_fingerprints.py

- ./utils/update_tests

- edit pierky/arouteserver/version.py

    - dev releases: vX.YY.Z-alpha1, vX.YY.Z-alpha2
    - prod release: vX.YY.Z

- edit CHANGES.rst

- verify RST syntax is ok
    python setup.py --long-description | rst2html.py --strict

- build and verify docs
    cd docs ; make html ; python -m SimpleHTTPServer 8000 ; cd ..

- new files to be added to MANIFEST.in?

- python setup.py sdist

- ~$ ./arouteserver/utils/test_new_rel

dev releases (in 'dev' branch):

    - git commit -a -m "v$(python -c 'from pierky.arouteserver.version import __version__; print(__version__)')"

    - git tag v$(python -c "from pierky.arouteserver.version import __version__; print(__version__)")

    - git push origin dev --tags

prod releases (in 'master' branch):

    - git commit -a -m "v$(python -c 'from pierky.arouteserver.version import __version__; print(__version__)')"

    - git tag v$(python -c "from pierky.arouteserver.version import __version__; print(__version__)")

    - git push origin master --tags

# upload to PyPi done by CD tools in GitHub/Travis
#- ~/.local/bin/twine upload dist/*
#- git push

- edit new release on GitHub
"""

__version__ = None

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Get proper long description for package
current_dir = dirname(abspath(__file__))
description = open(join(current_dir, "README.rst")).read()
changes = open(join(current_dir, "CHANGES.rst")).read()
long_description = '\n\n'.join([description, changes])
exec(open(join(current_dir, "pierky/arouteserver/version.py")).read())

install_requires = []
with open("requirements.txt", "r") as f:
    for line in f.read().split("\n"):
        if line:
            install_requires.append(line)

# Get the long description from README.md
setup(
    name="arouteserver",
    version=__version__,

    packages=["pierky", "pierky.arouteserver"],
    namespace_packages=["pierky"],
    package_data={
        "pierky.arouteserver": ["pierky/arouteserver/config.d/*",
                                "pierky/arouteserver/templates/*",
                                "pierky/arouteserver/tests/live_tests/skeleton/*.yml",
                                "pierky/arouteserver/tests/live_tests/skeleton/*.j2"]
    },
    include_package_data=True,
    
    license="GPLv3",
    description="A Python tool to automatically build (and test) configurations for BGP route servers.",
    long_description=long_description,
    url="https://github.com/pierky/arouteserver",
    download_url="https://github.com/pierky/arouteserver",

    author="Pier Carlo Chiodi",
    author_email="pierky@pierky.com",
    maintainer="Pier Carlo Chiodi",
    maintainer_email="pierky@pierky.com",

    install_requires=install_requires,
    tests_require=[
        "nose",
        "mock",
    ],
    test_suite="nose.collector",

    scripts=["scripts/arouteserver"],

    keywords=['BGP', 'Route server', 'BIRD', 'IP Routing'],

    classifiers=[
        "Development Status :: 4 - Beta",

        "Environment :: Console",

        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",

        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

        "Operating System :: POSIX",
        "Operating System :: Unix",

        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",

        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Networking",
    ],
)



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
import logging

from .base import ConfigParserBase, convert_deprecated
from .validators import *
from ..errors import ConfigError, ARouteServerError


class ConfigParserClients(ConfigParserBase):

    ROOT = "clients"

    def __init__(self, general_cfg=None):
        ConfigParserBase.__init__(self)
        self.general_cfg = general_cfg

    def parse(self):

        def get_client_descr(client):
            client_descr = ""
            if "asn" in client:
                client_descr += "AS{}".format(client["asn"])
            if "ip" in client:
                client_descr += " " + client["ip"]
            if not client_descr:
                client_descr = "unknown client"
            return client_descr

        if "clients" not in self.cfg:
            raise ConfigError("Missing top 'clients' statement.")
        if "asns" in self.cfg:
            del self.cfg["asns"]

        errors = False

        schema = {
            "asn": ValidatorASN(),
            "ip": ValidatorIPAddr(),
            "description": ValidatorText(mandatory=False),
            "password": ValidatorText(mandatory=False),
            "cfg": {
                "prepend_rs_as": ValidatorBool(mandatory=False),
                "passive": ValidatorBool(mandatory=False),
                "gtsm": ValidatorBool(mandatory=False),
                "add_path": ValidatorBool(mandatory=False),
                "filtering": {
                    "next_hop": {
                        "policy": ValidatorOption("policy",
                                                  ("strict", "same-as",
                                                   "authorized_addresses"),
                                                  mandatory=False),
                        "authorized_addresses_list": ValidatorListOf(
                            ValidatorIPAddr, mandatory=False
                        )
                    },
                    "ipv4_pref_len": ValidatorIPMinMaxLen(4, mandatory=False),
                    "ipv6_pref_len": ValidatorIPMinMaxLen(6, mandatory=False),
                    "max_as_path_len": ValidatorMaxASPathLen(mandatory=False),
                    "reject_invalid_as_in_as_path": ValidatorBool(mandatory=False),
                    "irrdb": {
                        "as_sets": ValidatorListOf(ValidatorASSet,
                                                   mandatory=False),
                        "enforce_origin_in_as_set": ValidatorBool(mandatory=False),
                        "enforce_prefix_in_as_set": ValidatorBool(mandatory=False),
                        "white_list_pref": ValidatorListOf(
                            ValidatorPrefixListEntry, mandatory=False,
                        ),
                        "white_list_asn": ValidatorASNList(mandatory=False),
                        "white_list_route": ValidatorListOf(
                            ValidatorWhiteListRouteEntry, mandatory=False
                        )
                    },
                    "rpki_bgp_origin_validation": {
                        "enabled": ValidatorBool(mandatory=False),
                        "reject_invalid": ValidatorBool(mandatory=False),
                    },
                    "black_list_pref": ValidatorListOf(
                        ValidatorPrefixListEntry, mandatory=False,
                    ),
                    "max_prefix": {
                        "action": ValidatorOption(
                            "action",
                            ("shutdown", "restart", "block", "warning"),
                            mandatory=False
                        ),
                        "restart_after": ValidatorUInt(mandatory=False),
                        "peering_db": {
                            "enabled": ValidatorBool(mandatory=False),
                            "increment": {
                                "absolute": ValidatorUInt(mandatory=False),
                                "relative": ValidatorUInt(mandatory=False)
                            }
                        },
                        "limit_ipv4": ValidatorUInt(mandatory=False),
                        "limit_ipv6": ValidatorUInt(mandatory=False),
                    },
                    "reject_policy": {
                        "policy": ValidatorOption("reject_policy",
                                                  ("reject", "tag"),
                                                  mandatory=False)
                    },
                },
                "blackhole_filtering" : {
                   "announce_to_client": ValidatorBool(mandatory=False),
                },
                "graceful_shutdown": {
                    "enabled": ValidatorBool(mandatory=False)
                },
                "attach_custom_communities": ValidatorListOf(ValidatorText,
                                                             mandatory=False)
            }
        }

        # Split configurations with more than one IP address into
        # multiple clients
        for client in self.cfg["clients"]:
            if "ip" in client:
                if isinstance(client["ip"], list):
                    for ip in client["ip"]:
                        client_clone = deepcopy(client)
                        client_clone["ip"] = ip
                        self.cfg["clients"].append(client_clone)
                    client["to_be_removed"] = True
        self.cfg["clients"] = [c for c in self.cfg["clients"] if "to_be_removed" not in c]

        # Clients' config validation
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)

            try:
                # Convert next_hop_policy (< v0.6.0) into the new format
                if "cfg" in client:
                    convert_deprecated(client["cfg"])

                ConfigParserBase.validate(schema, client, "clients")
            except ARouteServerError as e:
                err_msg = ("One or more errors occurred while processing "
                           "the client configuration for "
                           "'{}'".format(client_descr))
                if str(e):
                    err_msg += ": " + str(e)
                logging.error(err_msg)
                raise ConfigError()

        def inherit_from_general_cfg(dest, src, schema):
            for k in schema:
                if isinstance(schema[k], dict):
                    if k in src:
                        inherit_from_general_cfg(dest[k], src[k], schema[k])
                else:
                    if k not in dest or dest[k] is None:
                        if k in src:
                            dest[k] = src[k]

        # Inherit missing options from the general configuration.
        if self.general_cfg:
            for client in self.cfg["clients"]:
                inherit_from_general_cfg(client["cfg"], self.general_cfg, schema["cfg"])

        # Duplicate IP addresses?
        unique_ip = []
        for client in self.cfg["clients"]:
            ip = client["ip"]
            if ip in unique_ip:
                logging.error(
                    "Duplicate IP address found: {}.".format(ip)
                )
                errors = True
            else:
                unique_ip.append(ip)

        # Clients with...
        # - next_hop.policy == "authorized_addresses" AND
        # - no authorized IP addresses
        # ... or with...
        # - "authorized_addresses_list" AND
        # - next_hop.policy != "authorized_addresses"
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)
            next_hop = client["cfg"]["filtering"]["next_hop"]

            if next_hop["policy"] == "authorized_addresses" and \
                not next_hop["authorized_addresses_list"]:

                logging.error("The next_hop policy for client {} "
                              "is set to 'authorized_addresses' but "
                              "the list of authorized IP addresses "
                              "('authorized_addresses_list') is empty".format(
                                  client_descr
                                ))
                errors = True

            if next_hop["policy"] != "authorized_addresses" and \
                "authorized_addresses_list" in next_hop and \
                next_hop["authorized_addresses_list"] is not None:

                logging.error("The next_hop policy for client {} "
                              "is not 'authorized_addresses' but "
                              "the 'authorized_addresses_list' option "
                              "is set".format(
                                  client_descr
                                ))
                errors = True

        # Custom BGP communities must be declared within the general cfg
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)

            if not client["cfg"]["attach_custom_communities"]:
                continue

            for comm in client["cfg"]["attach_custom_communities"]:
                if self.general_cfg and \
                    comm not in self.general_cfg["custom_communities"]:

                    logging.error("The custom BGP community {} "
                                  "referenced on client {} is not declared on "
                                  "the general configuration.".format(
                                    comm, client_descr
                                    ))
                    errors = True

        if errors:
            raise ConfigError()



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import ipaddr
    ip_library = "ipaddr"
except ImportError:
    import ipaddress
    ip_library = "ipaddress"


class IPAddress(object):

    def __init__(self, ip):
        if ip_library == "ipaddr":
            self.obj = ipaddr.IPAddress(ip)
            self.ip = str(self.obj)
        else:
            self.obj = ipaddress.ip_address(ip)
            self.ip = str(self.obj.compressed)

        self.version = self.obj.version

    def __str__(self):
        return self.ip


class IPNetwork(object):

    def __init__(self, ip):
        if ip_library == "ipaddr":
            self.obj = ipaddr.IPNetwork(ip)
            self.ip = str(self.obj.ip)
        else:
            self.obj = ipaddress.ip_network(ip)
            self.ip = str(self.obj.network_address)

        self.prefixlen = self.obj.prefixlen
        self.max_prefixlen = self.obj.max_prefixlen
        self.version = self.obj.version

    def __str__(self):
        return "{}/{}".format(self.ip, self.prefixlen)
